# -*- coding: utf-8 -*-
"""
    wakatime.main
    ~~~~~~~~~~~~~

    Module entry point.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import print_function

import logging
import os
import sys
import time
import traceback

pwd = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.dirname(pwd))
sys.path.insert(0, os.path.join(pwd, 'packages'))

from .__about__ import __version__
from .api import send_heartbeats, get_time_today
from .arguments import parse_arguments
from .compat import u, json
from .constants import SUCCESS, UNKNOWN_ERROR, HEARTBEATS_PER_REQUEST
from .logger import setup_logging

log = logging.getLogger('WakaTime')

from .heartbeat import Heartbeat
from .offlinequeue import Queue


def execute(argv=None):
    if argv:
        sys.argv = ['wakatime'] + argv

    try:
        args, configs = parse_arguments()
    except SystemExit as ex:
        return ex.code

    setup_logging(args, __version__)

    if args.today:
        text, retval = get_time_today(args)
        if text:
            print(text)
        return retval

    try:
        heartbeats = []

        hb = Heartbeat(vars(args), args, configs)
        if hb:
            heartbeats.append(hb)
        elif args.entity:
            log.debug(hb.skip)

        if args.extra_heartbeats:
            try:
                for extra_data in json.loads(sys.stdin.readline()):
                    hb = Heartbeat(extra_data, args, configs)
                    if hb:
                        heartbeats.append(hb)
                    else:
                        log.debug(hb.skip)
            except json.JSONDecodeError as ex:
                log.warning(u('Malformed extra heartbeats json: {msg}').format(
                    msg=u(ex),
                ))

        retval = SUCCESS
        while heartbeats:
            retval = send_heartbeats(heartbeats[:HEARTBEATS_PER_REQUEST], args, configs)
            heartbeats = heartbeats[HEARTBEATS_PER_REQUEST:]
            if retval != SUCCESS:
                break

        if heartbeats:
            Queue(args, configs).push_many(heartbeats)

        if retval == SUCCESS:
            queue = Queue(args, configs)
            for offline_heartbeats in queue.pop_many(args.sync_offline_activity):
                time.sleep(1)
                retval = send_heartbeats(offline_heartbeats, args, configs)
                if retval != SUCCESS:
                    break

        return retval

    except:
        log.traceback(logging.ERROR)
        print(traceback.format_exc())
        return UNKNOWN_ERROR



# -*- coding: utf-8 -*-
"""
    wakatime.heartbeat
    ~~~~~~~~~~~~~~~~~~
    :copyright: (c) 2017 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""


import os
import logging
import re
from subprocess import PIPE

from .compat import u, json, is_win, Popen
from .exceptions import SkipHeartbeat
from .project import get_project_info
from .stats import get_file_stats
from .utils import get_user_agent, should_exclude, format_file_path, find_project_file


log = logging.getLogger('WakaTime')


class Heartbeat(object):
    """Heartbeat data for sending to API or storing in offline cache."""

    skip = False
    args = None
    configs = None

    time = None
    entity = None
    type = None
    category = None
    is_write = None
    project = None
    branch = None
    language = None
    dependencies = None
    lines = None
    lineno = None
    cursorpos = None
    user_agent = None

    _sensitive_when_hiding_filename = (
        'dependencies',
        'lines',
        'lineno',
        'cursorpos',
    )
    _sensitive_when_hiding_branch = (
        'branch',
    )

    def __init__(self, data, args, configs, _clone=None):
        if not data:
            self.skip = u('Skipping because heartbeat data is missing.')
            return

        self.args = args
        self.configs = configs

        self.entity = data.get('entity')
        self.time = data.get('time', data.get('timestamp'))
        self.is_write = data.get('is_write')
        self.user_agent = data.get('user_agent') or get_user_agent(args.plugin)

        self.type = data.get('type', data.get('entity_type'))
        if self.type not in ['file', 'domain', 'app']:
            self.type = 'file'

        self.category = data.get('category')
        allowed_categories = [
            'coding',
            'building',
            'indexing',
            'debugging',
            'running tests',
            'manual testing',
            'writing tests',
            'browsing',
            'code reviewing',
            'designing',
        ]
        if self.category not in allowed_categories:
            self.category = None

        if not _clone:
            exclude = self._excluded_by_pattern()
            if exclude:
                self.skip = u('Skipping because matches exclude pattern: {pattern}').format(
                    pattern=u(exclude),
                )
                return
            if self.type == 'file':
                self.entity = format_file_path(self.entity)
                self._format_local_file()
                if not self._file_exists():
                    self.skip = u('File does not exist; ignoring this heartbeat.')
                    return
                if self._excluded_by_missing_project_file():
                    self.skip = u('Skipping because missing .wakatime-project file in parent path.')
                    return

            if args.local_file and not os.path.isfile(args.local_file):
                args.local_file = None

            project, branch = get_project_info(configs, self, data)
            self.project = project
            self.branch = branch

            if self._excluded_by_unknown_project():
                self.skip = u('Skipping because project unknown.')
                return

            try:
                stats = get_file_stats(self.entity,
                                       entity_type=self.type,
                                       lineno=data.get('lineno'),
                                       cursorpos=data.get('cursorpos'),
                                       plugin=args.plugin,
                                       language=data.get('language'),
                                       local_file=args.local_file)
            except SkipHeartbeat as ex:
                self.skip = u(ex) or 'Skipping'
                return

        else:
            self.project = data.get('project')
            self.branch = data.get('branch')
            stats = data

        for key in ['language', 'dependencies', 'lines', 'lineno', 'cursorpos']:
            if stats.get(key) is not None:
                setattr(self, key, stats[key])

    def update(self, attrs):
        """Return a copy of the current Heartbeat with updated attributes."""

        data = self.dict()
        data.update(attrs)
        heartbeat = Heartbeat(data, self.args, self.configs, _clone=True)
        return heartbeat

    def sanitize(self):
        """Removes sensitive data including file names and dependencies.

        Returns a Heartbeat.
        """

        if self.entity is None:
            return self

        if self.type != 'file':
            return self

        if self._should_obfuscate_filename():
            self._sanitize_metadata(keys=self._sensitive_when_hiding_filename)
            if self._should_obfuscate_branch(default=True):
                self._sanitize_metadata(keys=self._sensitive_when_hiding_branch)
            extension = u(os.path.splitext(self.entity)[1])
            self.entity = u('HIDDEN{0}').format(extension)
        elif self.should_obfuscate_project():
            self._sanitize_metadata(keys=self._sensitive_when_hiding_filename)
            if self._should_obfuscate_branch(default=True):
                self._sanitize_metadata(keys=self._sensitive_when_hiding_branch)
        elif self._should_obfuscate_branch():
            self._sanitize_metadata(keys=self._sensitive_when_hiding_branch)

        return self

    def json(self):
        return json.dumps(self.dict())

    def dict(self):
        return {
            'time': self.time,
            'entity': self._unicode(self.entity),
            'type': self.type,
            'category': self.category,
            'is_write': self.is_write,
            'project': self._unicode(self.project),
            'branch': self._unicode(self.branch),
            'language': self._unicode(self.language),
            'dependencies': self._unicode_list(self.dependencies),
            'lines': self.lines,
            'lineno': self.lineno,
            'cursorpos': self.cursorpos,
            'user_agent': self._unicode(self.user_agent),
        }

    def items(self):
        return self.dict().items()

    def get_id(self):
        return u('{time}-{type}-{category}-{project}-{branch}-{entity}-{is_write}').format(
            time=self.time,
            type=self.type,
            category=self.category,
            project=self._unicode(self.project),
            branch=self._unicode(self.branch),
            entity=self._unicode(self.entity),
            is_write=self.is_write,
        )

    def should_obfuscate_project(self):
        """Returns True if hide_project_names is true or the entity file path
        matches one in the list of obfuscated project paths."""

        for pattern in self.args.hide_project_names:
            try:
                compiled = re.compile(pattern, re.IGNORECASE)
                if compiled.search(self.entity):
                    return True
            except re.error as ex:
                log.warning(u('Regex error ({msg}) for hide_project_names pattern: {pattern}').format(
                    msg=u(ex),
                    pattern=u(pattern),
                ))

        return False

    def _should_obfuscate_filename(self):
        """Returns True if hide_file_names is true or the entity file path
        matches one in the list of obfuscated file paths."""

        for pattern in self.args.hide_file_names:
            try:
                compiled = re.compile(pattern, re.IGNORECASE)
                if compiled.search(self.entity):
                    return True
            except re.error as ex:
                log.warning(u('Regex error ({msg}) for hide_file_names pattern: {pattern}').format(
                    msg=u(ex),
                    pattern=u(pattern),
                ))

        return False

    def _should_obfuscate_branch(self, default=False):
        """Returns True if hide_file_names is true or the entity file path
        matches one in the list of obfuscated file paths."""

        # when project names or file names are hidden and hide_branch_names is
        # not set, we default to hiding branch names along with file/project.
        if default and self.args.hide_branch_names is None:
            return True

        if not self.branch or not self.args.hide_branch_names:
            return False

        for pattern in self.args.hide_branch_names:
            try:
                compiled = re.compile(pattern, re.IGNORECASE)
                if compiled.search(self.entity) or compiled.search(self.branch):
                    return True
            except re.error as ex:
                log.warning(u('Regex error ({msg}) for hide_branch_names pattern: {pattern}').format(
                    msg=u(ex),
                    pattern=u(pattern),
                ))

        return False

    def _unicode(self, value):
        if value is None:
            return None
        return u(value)

    def _unicode_list(self, values):
        if values is None:
            return None
        return [self._unicode(value) for value in values]

    def _file_exists(self):
        return (self.entity and os.path.isfile(self.entity) or
            self.args.local_file and os.path.isfile(self.args.local_file))

    def _format_local_file(self):
        """When args.local_file empty on Windows, tries to map args.entity to a
        unc path.

        Updates args.local_file in-place without returning anything.
        """

        if self.type != 'file':
            return

        if not self.entity:
            return

        if not is_win:
            return

        if self._file_exists():
            return

        self.args.local_file = self._to_unc_path(self.entity)

    def _to_unc_path(self, filepath):
        drive, rest = self._splitdrive(filepath)
        if not drive:
            return filepath

        stdout = None
        try:
            stdout, stderr = Popen(['net', 'use'], stdout=PIPE, stderr=PIPE).communicate()
        except OSError:
            pass
        else:
            if stdout:
                cols = None
                for line in stdout.strip().splitlines()[1:]:
                    line = u(line)
                    if not line.strip():
                        continue
                    if not cols:
                        cols = self._unc_columns(line)
                        continue
                    start, end = cols.get('local', (0, 0))
                    if not start and not end:
                        break
                    local = line[start:end].strip().split(':')[0].upper()
                    if not local.isalpha():
                        continue
                    if local == drive:
                        start, end = cols.get('remote', (0, 0))
                        if not start and not end:
                            break
                        remote = line[start:end].strip()
                        return remote + rest

        return filepath

    def _unc_columns(self, line):
        cols = {}
        current_col = u('')
        newcol = False
        start, end = 0, 0
        for char in line:
            if char.isalpha():
                if newcol:
                    cols[current_col.strip().lower()] = (start, end)
                    current_col = u('')
                    start = end
                    newcol = False
                current_col += u(char)
            else:
                newcol = True
            end += 1
        if start != end and current_col:
            cols[current_col.strip().lower()] = (start, -1)
        return cols

    def _splitdrive(self, filepath):
        if filepath[1:2] != ':' or not filepath[0].isalpha():
            return None, filepath
        return filepath[0].upper(), filepath[2:]

    def _excluded_by_pattern(self):
        return should_exclude(self.entity, self.args.include, self.args.exclude)

    def _excluded_by_unknown_project(self):
        if self.project:
            return False
        return self.args.exclude_unknown_project

    def _excluded_by_missing_project_file(self):
        if not self.args.include_only_with_project_file:
            return False
        return find_project_file(self.entity) is None

    def _sanitize_metadata(self, keys=[]):
        for key in keys:
            setattr(self, key, None)

    def __repr__(self):
        return self.json()

    def __bool__(self):
        return not self.skip

    def __nonzero__(self):
        return self.__bool__()

    def __getitem__(self, key):
        return self.dict()[key]



# -*- coding: utf-8 -*-
"""
    wakatime.dependencies.javascript
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from JavaScript code.

    :copyright: (c) 2018 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import re

from . import TokenParser


class JavascriptParser(TokenParser):
    state = None
    extension = re.compile(r'\.\w{1,4}$')

    def parse(self):
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_token(self, token, content):
        if self.partial(token) == 'Reserved':
            self._process_reserved(token, content)
        elif self.partial(token) == 'Single':
            self._process_string(token, content)
        elif self.partial(token) == 'Punctuation':
            self._process_punctuation(token, content)
        else:
            self._process_other(token, content)

    def _process_reserved(self, token, content):
        if self.state is None:
            self.state = content

    def _process_string(self, token, content):
        if self.state == 'import':
            self.append(self._format_module(content))
        self.state = None

    def _process_punctuation(self, token, content):
        if content == ';':
            self.state = None

    def _process_other(self, token, content):
        pass

    def _format_module(self, content):
        content = content.strip().strip('"').strip("'").strip()
        content = content.split('/')[-1].split('\\')[-1]
        content = self.extension.sub('', content, count=1)
        return content


class TypeScriptParser(JavascriptParser):
    pass



# -*- coding: utf-8 -*-
"""
    wakatime.logger
    ~~~~~~~~~~~~~~~

    Provides the configured logger for writing JSON to the log file.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import logging
import os
import traceback

from .compat import u
from .packages.requests.packages import urllib3
try:
    from collections import OrderedDict  # pragma: nocover
except ImportError:  # pragma: nocover
    from .packages.ordereddict import OrderedDict
try:
    from .packages import simplejson as json  # pragma: nocover
except (ImportError, SyntaxError):  # pragma: nocover
    import json


class JsonFormatter(logging.Formatter):

    def setup(self, timestamp, is_write, entity, version, plugin, verbose,
              warnings=False):
        self.timestamp = timestamp
        self.is_write = is_write
        self.entity = entity
        self.version = version
        self.plugin = plugin
        self.verbose = verbose
        self.warnings = warnings

    def format(self, record, *args):
        data = OrderedDict([
            ('now', self.formatTime(record, self.datefmt)),
        ])
        data['version'] = u(self.version)
        if self.plugin:
            data['plugin'] = u(self.plugin)
        data['time'] = self.timestamp
        if self.verbose:
            data['caller'] = u(record.pathname)
            data['lineno'] = record.lineno
            if self.is_write:
                data['is_write'] = self.is_write
            data['file'] = u(self.entity)
        data['level'] = record.levelname
        data['message'] = u(record.getMessage() if self.warnings else record.msg)
        return json.dumps(data)

    def traceback(self, lvl=None):
        logger = logging.getLogger('WakaTime')
        if not lvl:
            lvl = logger.getEffectiveLevel()
        logger.log(lvl, traceback.format_exc())


def set_log_level(logger, args):
    level = logging.WARN
    if args.verbose:
        level = logging.DEBUG
    logger.setLevel(level)


def setup_logging(args, version):
    urllib3.disable_warnings()
    logger = logging.getLogger('WakaTime')
    for handler in logger.handlers:
        logger.removeHandler(handler)
    set_log_level(logger, args)
    logfile = args.log_file
    if not logfile:
        logfile = '~/.wakatime.log'
    handler = logging.FileHandler(os.path.expanduser(logfile))
    formatter = JsonFormatter(datefmt='%Y/%m/%d %H:%M:%S %z')
    formatter.setup(
        timestamp=args.timestamp,
        is_write=args.is_write,
        entity=args.entity,
        version=version,
        plugin=args.plugin,
        verbose=args.verbose,
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # add custom traceback logging method
    logger.traceback = formatter.traceback

    warnings_formatter = JsonFormatter(datefmt='%Y/%m/%d %H:%M:%S %z')
    warnings_formatter.setup(
        timestamp=args.timestamp,
        is_write=args.is_write,
        entity=args.entity,
        version=version,
        plugin=args.plugin,
        verbose=args.verbose,
        warnings=True,
    )
    warnings_handler = logging.FileHandler(os.path.expanduser(logfile))
    warnings_handler.setFormatter(warnings_formatter)
    logging.getLogger('py.warnings').addHandler(warnings_handler)
    try:
        logging.captureWarnings(True)
    except AttributeError:  # pragma: nocover
        pass  # Python >= 2.7 is needed to capture warnings

    return logger



# -*- coding: utf-8 -*-
"""
    wakatime.dependencies.rust
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from Rust code.

    :copyright: (c) 2018 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

from . import TokenParser


class RustParser(TokenParser):
    state = None

    def parse(self):
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_token(self, token, content):
        if self.partial(token) == 'Keyword':
            self._process_keyword(token, content)
        elif self.partial(token) == 'Whitespace':
            self._process_whitespace(token, content)
        elif self.partial(token) == 'Name':
            self._process_name(token, content)
        else:
            self._process_other(token, content)

    def _process_keyword(self, token, content):
        if self.state == 'extern' and content == 'crate':
            self.state = 'extern crate'
        else:
            self.state = content

    def _process_whitespace(self, token, content):
        pass

    def _process_name(self, token, content):
        if self.state == 'extern crate':
            self.append(content)
        self.state = None

    def _process_other(self, token, content):
        self.state = None
